// Persistent binary relationship for Netscape IFC

// Copyright (C) 1999 by Makoto Nagata, MicroNova <mnagata@micronova.com>.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//
// Visit http://www.micronova.com for updates

package com.micronova.util;

import netscape.util.*;

/**

   A Relation is a set of (source, target) pairs of Objects.
   Object comparison is identity-based, not 'equals'.  

   * @version 1.0
   * @author Makoto Nagata, MicroNova

*/

public class Relation implements Codable
{
    /**

       Hashtable that maps a source to its target vector.
       I.e., fore.get(source) is the vector containing all the
       target objects for a given source.

    */

    protected Hashtable fore;

    /**

       Hashtable that maps a target to its source vector.
       This is null by default, but automatically 
       built by <a href="#buildBackTable">buildBackTable()</a>
       when either <a href="#getTargets">getTargets()</a>
       or <a href="#getSourcesFor">getSourcesFor()</a>
       is performed.
       Once built, back.get(target) is always the vector
       containing all the source objects for a given target.
       
    */

    protected Hashtable back;

    /**

       Constructs an empty relation.  By default, the backward
       table is null.

    */

    public Relation()
    {
        fore = new Hashtable();
        back = null;
    }   

    /**

       Builds the backward hashtable.  This is called by
       <a href="#getTargets">getTargets()</a> or 
       <a href="#getSourcesFor">getSourcesFor()</a> 
       if the backward  hashtable is not built yet.

    */

    protected void buildBackTable()
    {
        back = new Hashtable();

        Enumeration sources = getSources().elements();

        while (sources.hasMoreElements())
        {
            Object source = sources.nextElement();

            Enumeration targets = getTargetsFor(source).elements();

            while (targets.hasMoreElements())
            {
                addPair(back, targets.nextElement(), source);
            }
        }
    }

    /**

       Returns the vector for a given object x from a given table.
       Returns null if x is not found.

    */

    protected Vector getVector(Hashtable table, Object x)
    {
        return (Vector)table.get(x);
    }

    /**

       Adds a new pair (x, y) to a table.
       I.e., y is added to the vector for x
       if it is not present (comparison is identity-based).

    */

    protected void addPair(Hashtable table, Object x, Object y)
    {
        Vector list = getVector(table, x);

        if (list == null)
        {
            list = new Vector();

            table.put(x, list);
        }

        if (!list.containsIdentical(y))
        {
            list.addElement(y);
        }
    }

    /**

       Removes a pair (x, y) from a table.
       If doCleanup is true, the vector for x is removed from 
       the hashtable if it becomes empty after removal.  
       Otherwise, the vector stays in the hashtable.

    */

    protected void removePair(Hashtable table, Object x, Object y, boolean doCleanup)
    {
        Vector list = getVector(table, x);

        if (list != null)
        {
            list.removeElement(y);

            if (doCleanup)
            {
                if (list.isEmpty())
                {
                    table.remove(x);
                }
            }
        }
    }

    /**

       Returns all the source objects in the set.
       Always returns a non-null (cloned) Vector.

    */

    public Vector getSources()
    {
        return fore.keysVector();
    }

    /**

       Returns all the target objects in the set.  
       Always returns a non-null (cloned) Vector.
       The backward table is built if not present by
       <a href="#buildBackTable">buildBackTable()</a>.

    */

    public Vector getTargets()
    {
        if (back == null)
        {
            buildBackTable();
        }

        return back.keysVector();
    }

    /**

       Places a new pair (source, target) in the relation set.

    */

    public void put(Object source, Object target)
    {
        addPair(fore, source, target);

        if (back != null)
        {
            addPair(back, target, source);
        }
    }

    /**

       Removes a pair (source, target) from the relation set.
       See <a href="#removePair">removePair()</a>.

    */

    public void remove(Object source, Object target, boolean doCleanup)
    {
        removePair(fore, source, target, doCleanup);

        if (back != null)
        {
            removePair(back, target, source, doCleanup);
        }
    }

    /**

       Removes a pair (source, target) from the relation set (with
       cleaning up).  Same as remove(source, target, true).

    */

    public void remove(Object source, Object target)
    {
        remove(source, target, true);
    }

    /**

      Returns all the targets for a given source.
      Always returns a non-null Vector.  If the
      returned vector is non-empty, it is the 
      actual vector stored in the hashtable.
      <b>Do not modify the returned vector.</b>

    */

    public Vector getTargetsFor(Object source)
    {
        Vector v = getVector(fore, source);

        if (v == null)
        {
            v = new Vector();
        }

        return v;
    }

    /**

       Returns all the sources for a given target.
       
       The backward table is built if not present.

       Always returns a non-null Vector.  If the
       returned vector is non-empty, it is the 
       actual vector stored in the hashtable.
       <b>Do not modify the returned vector.</b>

    */

    public Vector getSourcesFor(Object target)
    {
        if (back == null)
        {
            buildBackTable();
        }

        Vector v = getVector(back, target);

        if (v == null)
        {
            v = new Vector();
        }

        return v;
    }

    /**

       Returns true if the given (source, target) pair exists
       in the relation set.  Comparison is identity-based,
       not 'equals'.

    */

    public boolean contains(Object source, Object target)
    {
        Vector v = getVector(fore, source);

        if (v == null)
        {
            return false;
        }
        else
        {
            return v.containsIdentical(target);
        }
    }

    /**

       Removes all pairs (*, x) and (x, *) from the relation set.
       See <a href="#removePair">removePair()</a>.
    */

    public void remove(Object x, boolean doCleanup)
    {
        if (back != null)
        {
            Enumeration targets = ((Vector)getTargetsFor(x).clone()).elements();
            while (targets.hasMoreElements())
            {
                remove(x, targets.nextElement(), doCleanup);
            }

            Enumeration sources = ((Vector)getSourcesFor(x).clone()).elements();

            while (sources.hasMoreElements())
            {
                remove(sources.nextElement(), x, doCleanup);
            }
        }
        else
        {
            Enumeration sources = getSources().elements();

            while (sources.hasMoreElements())
            {
                Object source = sources.nextElement();

                Enumeration targets = ((Vector)getTargetsFor(source).clone()).elements();

                if (x == source)
                {
                    while (targets.hasMoreElements())
                    {
                        remove(source, targets.nextElement(), doCleanup);
                    }
                }
                else
                {
                    while (targets.hasMoreElements())
                    {
                        Object target = targets.nextElement();

                        if (target == x)
                        {
                            remove(source, x, doCleanup);
                        }
                    }
                }
            }
        }
    }

    /**

       Removes all pairs (*, x) and (x, *) from the relation set (with cleaning up).
       Same as remove(x, true).

    */

    public void remove(Object x)
    {
        remove(x, true);
    }

    /**

       Returns true if the backward table is built.

    */

    public boolean hasBackTable()
    {
        return (back != null);
    }

    // Codable interface

    public void describeClassInfo(ClassInfo info)
    {
        info.addClass(getClass().getName(), 1);
        info.addField("fore", OBJECT_TYPE);
    }

    public void decode(Decoder decoder) throws CodingException
    {
        fore = (Hashtable)decoder.decodeObject("fore");
    }

    public void encode(Encoder encoder) throws CodingException
    {
        encoder.encodeObject("fore", fore);
    }

    public void finishDecoding()
    {
    }
}

